<!DOCTYPE html>
<html>
  <head>
    <title>Folder contents.</title>
    <meta charset="utf-8" />
    <meta content='text/html; charset=utf-8' http-equiv='Content-Type'>
    <style>
        dt {{
            font-weight: bold;
        }}
    </style>
  </head>
  <body>
    <h1>&#47;{path}</h1>
    <ul>{info}</ul>
  </body>
</html>